# Copyright 2024 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""Useful functions for uploading checkpoints to remote object store."""

import logging
import os
import pathlib
import time
from concurrent.futures import Future
from typing import Optional

import composer.utils.dist as dist
from composer.core import Callback, State
from composer.loggers import Logger, MosaicMLLogger
from composer.utils import (
    RemoteFilesExistingCheckStatus,
    RemoteUploader,
    create_symlink_file,
    parse_uri,
)

log = logging.getLogger(__name__)


class CheckpointUploadCallback(Callback):
    """Callback class for async checkpoint uploading."""

    def __init__(
        self,
        remote_uploader: RemoteUploader,
        upload_timeout_in_seconds: int = 3600,
    ):
        self.remote_uploader = remote_uploader
        if dist.get_global_rank() == 0:
            self.rank_saves_symlinks = True
        else:
            self.rank_saves_symlinks = False

        # Each Tuple contains information of:
        # the future of checking if checkpoint files upload finish
        # local symlink file name
        # remote symlink file name
        self.symlink_upload_tasks: list[tuple[Future[RemoteFilesExistingCheckStatus], str, Optional[str]]] = []
        self.upload_timeout_in_seconds: int = upload_timeout_in_seconds

        # Allow unit test to override this to make it faster
        self._symlink_upload_wait_before_next_try_in_seconds = 30.0

    def add_symlink_upload_task(
        self,
        remote_file_names: list[str],
        symlink_file_name: Optional[str] = None,
        symlink_remote_file_name: Optional[str] = None,
    ):
        check_remote_files_exist_future = None
        if self.rank_saves_symlinks and symlink_file_name is not None:
            check_remote_files_exist_future = self.remote_uploader.check_remote_files_exist_async(
                remote_checkpoint_file_names=remote_file_names,
                max_wait_time_in_seconds=self.upload_timeout_in_seconds,
                wait_before_next_try_in_seconds=self._symlink_upload_wait_before_next_try_in_seconds,
            )
            self.symlink_upload_tasks.append(
                (check_remote_files_exist_future, symlink_file_name, symlink_remote_file_name),
            )

    def _log_checkpoint_upload(self, logger: Logger):
        for destination in logger.destinations:
            if isinstance(destination, MosaicMLLogger):
                destination.log_metadata({'checkpoint_uploaded_time': time.time()}, force_flush=True)

    def batch_end(self, state: State, logger: Logger) -> None:
        del state  # unused
        self.remote_uploader.check_workers()
        if not self.rank_saves_symlinks:
            return
        undone_symlink_upload_tasks = []
        for (check_remote_files_exist_future, local_symlink_file,
             remote_symlink_file) in reversed(self.symlink_upload_tasks):
            if not check_remote_files_exist_future.done():
                undone_symlink_upload_tasks.insert(
                    0,
                    (check_remote_files_exist_future, local_symlink_file, remote_symlink_file),
                )
                continue
            else:
                result = check_remote_files_exist_future.result()
                if result == RemoteFilesExistingCheckStatus.EXIST:
                    assert remote_symlink_file is not None
                    self.remote_uploader.upload_file_async(
                        remote_file_name=remote_symlink_file,
                        file_path=pathlib.Path(local_symlink_file),
                        overwrite=True,
                    )
                    self._log_checkpoint_upload(logger)
                    break
                else:
                    raise RuntimeError(f'Failed to check if checkpoint files upload finish: {result}')
        self.symlink_upload_tasks = undone_symlink_upload_tasks

    def wait(self) -> None:
        log.info('Waiting for checkpoint uploading to finish')
        self.remote_uploader.wait()
        if self.rank_saves_symlinks and len(self.symlink_upload_tasks) > 0:
            log.debug('Uploading symlink to the latest checkpoint')
            # We only need to upload a symlink pointing to the latest checkpoint files, so we can ignore successful uploads of older checkpoints.
            check_remote_files_exist_future, local_symlink_file, remote_symlink_file = self.symlink_upload_tasks[-1]
            result = check_remote_files_exist_future.result()
            if result == RemoteFilesExistingCheckStatus.EXIST:
                assert remote_symlink_file is not None
                symlink_upload_future = self.remote_uploader.upload_file_async(
                    remote_file_name=remote_symlink_file,
                    file_path=pathlib.Path(local_symlink_file),
                    overwrite=True,
                )
                symlink_upload_future.result()
            else:
                raise RuntimeError(f'Failed to check if checkpoint files upload finish: {result}')
            self.symlink_upload_tasks = []
        log.info('Checkpoint uploading finished!')

    def fit_end(self, state: State, logger: Logger) -> None:
        del state  # unused
        self.wait()
        self._log_checkpoint_upload(logger)

    def post_close(self):
        # Wait the symlink file upload to finish and close remote uploader
        try:
            self.remote_uploader.wait_and_close()
        except Exception as e:
            log.error(f'RemoteUploader run into exception {e}')



def upload_file(
    dest_dir: str,
    source_path: Optional[str]=None,
    symlink_granularity: Optional[str]=None, # file, dir, or None
    symlink_name: Optional[str]='latest.symlink',
    async_upload: bool = True,
    state: Optional[State] = None,
    overwrite: bool = False,
):
    """Standalone function for uploading a checkpoint file.

    This function does not actually upload the checkpoint; it initiates the RemoteUploader's uploading of it
    Args:
        source_path (str): The path to the file to upload.
        dest_dir (str): The directory/uri to upload the file to.
        symlink_granularity (Optional[str]): The granularity to use for symlinking. One of 'file', 'dir', or None.
            if None: no symlink uploaded
            if 'file': command remoteuploader to wait until the file (specificied by source_path) is uploaded and then uploads a symlink pointing to the uploaded file
            if 'dir': command remoteuploader  to wait until all files across all ranks are uploaded to dest_dir and then uploads a symlink
                pointing to the remote directory (prefix in object_store terminology).
        symlink_name (Optional[str]): The name to use for the symlink. Defaults to 'latest.symlink'.
        async_upload (bool): If True, the uploads will be done asynchronously via the RemoteUploader and this function will return immediately.
        state (Optional[State]): If async_upload is True, then state must be specified so that the remote_uploader can be
            either extracted from state.callbacks or initialized and added to state.callbacks.
        overwrite (bool): If allow overwrite existing remote checkpoint files
    """
    remote_uploader = RemoteUploader(remote_folder=dest_dir)
    dest_path = ''
    if source_path is not None:
        _, _, dest_path = parse_uri(dest_dir)
        remote_file_name = os.path.join(dest_path, os.path.basename(source_path))
    else:
        remote_file_name = None
    all_remote_file_names = dist.all_gather_object([remote_file_name] if remote_file_name is not None else [])
    remote_file_names = []
    for filenames in all_remote_file_names:
        remote_file_names += filenames

    if source_path is not None:
        assert remote_uploader is not None
        assert remote_file_name is not None
        remote_uploader.upload_file_async(
            remote_file_name=remote_file_name,
            file_path=pathlib.Path(source_path),
            overwrite=overwrite,
        )
    symlink_remote_file_name = None
    if dist.get_global_rank() == 0:
        if symlink_name is not None:
            symlink_remote_file_name = os.path.join(dest_path, symlink_name)
            if symlink_granularity == 'file':
                create_symlink_file(dest_path, symlink_name)
            elif symlink_granularity == 'dir':
                create_symlink_file(str(pathlib.Path(dest_path).parent), symlink_name)
            elif symlink_granularity is not None:
                raise ValueError(f'Unrecognized symlink granularity: {symlink_granularity}')
    else:
        symlink_remote_file_name = None
        symlink_name = None

    upload_callback = None
    if state is not None:
        for callback in state.callbacks:
            if isinstance(callback, CheckpointUploadCallback):
                upload_callback = callback
                break
    if upload_callback is None:
        upload_callback = CheckpointUploadCallback(remote_uploader=remote_uploader)
        if state is not None:
            state.callbacks.append(upload_callback)
    upload_callback.add_symlink_upload_task(
        remote_file_names=remote_file_names,
        symlink_file_name=symlink_name,
        symlink_remote_file_name=symlink_remote_file_name,
    )

    if not async_upload:
        upload_callback.wait()



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""A collection of common torchmetrics."""

from composer.metrics.map import MAP
from composer.metrics.metrics import CrossEntropy, Dice, LossMetric, MIoU
from composer.metrics.nlp import (
    BinaryF1Score,
    LanguageCrossEntropy,
    LanguagePerplexity,
    MaskedAccuracy,
)

__all__ = [
    'MAP',
    'MIoU',
    'Dice',
    'CrossEntropy',
    'LossMetric',
    'BinaryF1Score',
    'LanguageCrossEntropy',
    'MaskedAccuracy',
    'LanguagePerplexity',
]



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""Parallelism configs."""

from dataclasses import dataclass, field
from typing import Any, Optional

from torch.distributed._tensor.device_mesh import DeviceMesh


@dataclass
class FSDPConfig:
    """Configuration for Fully Sharded Data Parallelism (FSDP)."""
    activation_checkpointing: bool = False
    activation_checkpointing_reentrant: bool = True
    activation_cpu_offload: bool = False
    auto_wrap: bool = True
    te_checkpoint_wrapper: bool = False
    te_shard_fp8_weight: bool = False
    backward_prefetch: str = 'BACKWARD_POST'
    backward_prefetch_limit: int = 1
    cpu_offload: bool = False
    data_parallel_shard_degree: int = -1
    data_parallel_replicate_degree: Optional[int] = None
    forward_prefetch: bool = False
    forward_prefetch_limit: int = 1
    ignored_modules: Optional[Any] = None
    keep_low_precision_grads: bool = False
    limit_all_gathers: bool = True
    load_monolith_rank0_only: bool = False
    load_planner: Optional[Any] = None
    mixed_precision: str = 'DEFAULT'
    process_group: Optional[Any] = None
    save_planner: Optional[Any] = None
    sharded_ckpt_prefix_dir: str = 'ep{epoch}-ba{batch}'
    sharding_strategy: str = 'FULL_SHARD'
    state_dict_type: str = 'full'
    sync_module_states: bool = False
    use_orig_params: bool = True
    verbose: bool = False

    _device_mesh: Optional[DeviceMesh] = field(default=None, init=False, repr=False)

    def __init__(self, **kwargs):
        if 'device_mesh' in kwargs or '_device_mesh' in kwargs:
            raise ValueError(
                f'Directly specifying device mesh for FSDP was deprecated in Composer version 0.24.0. ' +
                f"Please specify 'data_parallel_shard_degree' and/or 'data_parallel_replicate_degree' instead.",
            )

        for k, v in kwargs.items():
            setattr(self, k, v)

    @property
    def device_mesh(self) -> Optional[DeviceMesh]:
        return self._device_mesh

    @device_mesh.setter
    def device_mesh(self, value: Optional[DeviceMesh]):
        self._device_mesh = value


@dataclass
class TPConfig:
    """Configuration for tensor parallelism (TP)."""
    device_mesh: Optional[DeviceMesh] = None
    tensor_parallel_degree: int = 1
    layer_plan: Any = None


@dataclass
class ParallelismConfig:
    """Configuration for parallelism."""
    fsdp: Optional[FSDPConfig] = None
    tp: Optional[TPConfig] = None



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""FSDP related configs and helper functions."""

import logging
import warnings
from typing import Any, Union

import torch
from torch import distributed
from torch.distributed import ProcessGroup
from torch.distributed.fsdp import (
    BackwardPrefetch,
    CPUOffload,
    MixedPrecision,
    ShardingStrategy,
)

from composer.core import Precision
from composer.utils import dist

log = logging.getLogger(__name__)

SHARDING_MAP = {
    'NO_SHARD': ShardingStrategy.NO_SHARD,
    'SHARD_GRAD_OP': ShardingStrategy.SHARD_GRAD_OP,
    'FULL_SHARD': ShardingStrategy.FULL_SHARD,
    '_HYBRID_SHARD_ZERO2': ShardingStrategy._HYBRID_SHARD_ZERO2,
    'HYBRID_SHARD': ShardingStrategy.HYBRID_SHARD,
}

BACKWARD_PREFETCH_MAP = {
    'NONE': None,
    'BACKWARD_PRE': BackwardPrefetch.BACKWARD_PRE,
    'BACKWARD_POST': BackwardPrefetch.BACKWARD_POST,
}


def _get_torch_dtype(dtype: Union[Precision, str]):
    """Convert common string representations of dtypes to torch dtypes."""
    dtype = dtype.value if isinstance(dtype, Precision) else dtype
    if dtype in ['float32', 'torch.float32', 'fp32']:
        return torch.float32
    elif dtype in ['float16', 'torch.float16', 'half', 'fp16', 'amp', 'amp_fp16']:
        return torch.float16
    elif dtype in ['bfloat16', 'bfloat', 'torch.bfloat16', 'bf16', 'amp_bf16']:
        return torch.bfloat16
    elif dtype in ['float8', 'torch.float8', 'fp8', 'amp_fp8']:
        if hasattr(torch, 'float8'):
            raise NotImplementedError('Torch has enabled float8. This should be updated to `return torch.float8`')
        else:
            warnings.warn('We use torch.bfloat16 by default for amp_fp8 as there is no fp8 datatype in PyTorch yet.')
            return torch.bfloat16
    else:
        raise ValueError(f'Not sure how to convert dtype={dtype} to a torch dtype.')


def get_mixed_precision(precision, mixed_precision='DEFAULT', keep_low_precision_grads=False):
    """Helper function for configuring mixed_precision."""
    param_dtype = None
    reduce_dtype = None
    buffer_dtype = None
    if isinstance(mixed_precision, dict):
        param_dtype = mixed_precision.get('param_dtype', None)
        if param_dtype is not None:
            param_dtype = _get_torch_dtype(param_dtype)
        reduce_dtype = mixed_precision.get('reduce_dtype', None)
        if reduce_dtype is not None:
            reduce_dtype = _get_torch_dtype(reduce_dtype)
        buffer_dtype = mixed_precision.get('buffer_dtype', None)
        if buffer_dtype is not None:
            buffer_dtype = _get_torch_dtype(buffer_dtype)
    elif isinstance(mixed_precision, str):
        mixed_precision = mixed_precision.upper()
        if mixed_precision == 'FULL':
            pass
        elif mixed_precision == 'DEFAULT':
            param_dtype = _get_torch_dtype(precision)
            reduce_dtype = torch.float32
            buffer_dtype = _get_torch_dtype(precision)
        elif mixed_precision == 'PURE':
            param_dtype = _get_torch_dtype(precision)
            reduce_dtype = _get_torch_dtype(precision)
            buffer_dtype = _get_torch_dtype(precision)
        else:
            raise ValueError(f'Unable to interpret mixed_precision={mixed_precision}')
    else:
        raise ValueError(f'Unable to interpret mixed_precision={mixed_precision}')

    mixed_precision = MixedPrecision(
        param_dtype=param_dtype,
        reduce_dtype=reduce_dtype,
        buffer_dtype=buffer_dtype,
        keep_low_precision_grads=keep_low_precision_grads,
    )

    return mixed_precision, param_dtype, reduce_dtype, buffer_dtype


def get_cpu_offload(cpu_offload=False):
    """Helper function for configuring cpu_offload."""
    cpu_offload = CPUOffload(offload_params=True) if cpu_offload else None
    if cpu_offload is not None:
        raise ValueError('FSDP CPU Offload not supported yet.')
    return cpu_offload


def _get_process_group(pg, process_group_cache=None):
    """Helper function for configuring and/or retrieving process groups."""
    if pg is None or isinstance(pg, ProcessGroup):  # Return as is, no caching
        return pg

    world_size = dist.get_world_size()
    local_world_size = dist.get_local_world_size()

    # Handle special str process_group cases
    if pg == 'self':
        pg = 'set1'
        log.info(f"Converting process_group='self' to process_group='{pg}'")
    elif pg == 'node':
        pg = f'set{local_world_size}'
        log.info(f"Converting process_group='node' to process_group='{pg}'")
    elif pg == 'local_rank_across_nodes':
        pg = f'mod{local_world_size}'
        log.info(f"Converting process_group='local_rank_across_nodes' to process_group='{pg}'")

    # Handle str and Union[list[int], tuple[int]] process_group cases
    if isinstance(pg, str) and pg.startswith('set'):
        k = int(pg.strip('set'))
        world_size = dist.get_world_size()
        if world_size % k != 0:
            raise RuntimeError(f'{world_size} must be divisible by set size ({k})')
        start = dist.get_global_rank() // k * k
        ranks = tuple(range(start, start + k))
    elif isinstance(pg, str) and pg.startswith('mod'):
        k = int(pg.strip('mod'))
        world_size = dist.get_world_size()
        if world_size % k != 0:
            raise RuntimeError(f'{world_size} must be divisible by mod ({k})')
        ranks = tuple(range(dist.get_global_rank() % k, world_size, k))
    elif isinstance(pg, (list, tuple)):
        ranks = tuple(pg)
    else:
        raise ValueError(f'Unsure how to setup process_group={pg}')

    if process_group_cache is not None and ranks in process_group_cache:
        log.info(f'Using cached progress group with {ranks=} on rank={dist.get_global_rank()}.')
        return process_group_cache[ranks]

    log.info(f'Instantiating custom process groups with {ranks=} on rank={dist.get_global_rank()}.')

    ranks_per_subgroup_list = list(set(dist.all_gather_object(ranks)))
    (
        current_group,
        _subgroups,
    ) = distributed.distributed_c10d.new_subgroups_by_enumeration(  # type: ignore[reportGeneralTypeIssues]
        ranks_per_subgroup_list,
    )

    if process_group_cache is not None:
        process_group_cache[ranks] = current_group
    return current_group


def set_custom_fsdp_module_kwargs(module_kwargs: dict, process_group_cache: dict[tuple[int], Any]) -> dict:
    """Set custom module_kwargs per fsdp module."""
    if ('sharding_strategy' in module_kwargs and module_kwargs['sharding_strategy'] not in SHARDING_MAP.values()):
        module_kwargs['sharding_strategy'] = SHARDING_MAP[module_kwargs['sharding_strategy'].upper()]
    if 'backward_prefetch' in module_kwargs:
        if module_kwargs['backward_prefetch'] not in BACKWARD_PREFETCH_MAP.values():
            module_kwargs['backward_prefetch'] = BACKWARD_PREFETCH_MAP[module_kwargs['backward_prefetch'].upper()]
    if 'cpu_offload' in module_kwargs and not isinstance(module_kwargs['cpu_offload'], CPUOffload):
        module_kwargs['cpu_offload'] = get_cpu_offload(cpu_offload=module_kwargs['cpu_offload'].upper())
    if 'mixed_precision' in module_kwargs and not isinstance(module_kwargs['mixed_precision'], MixedPrecision):
        # `precision` needs to set `'mixed_precision'`, but `precision` is not part of fsdp kwargs
        raise NotImplementedError(
            f"Automated setting of custom per module mixed_precision is not implemented, but it can be set if `isinstance(module_kwargs['mixed_precision'], MixedPrecision)`",
        )
    if 'process_group' in module_kwargs:
        # Call on every process group if it is a tuple/list of non-ints
        if type(module_kwargs['process_group']) in [
            list,
            tuple,
        ] and not all(isinstance(x, int) for x in module_kwargs['process_group']):
            module_kwargs['process_group'] = tuple(
                _get_process_group(pg, process_group_cache) for pg in module_kwargs['process_group']
            )
        else:
            module_kwargs['process_group'] = _get_process_group(module_kwargs['process_group'], process_group_cache)

    return module_kwargs



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""Util class to upload file."""

import logging
import multiprocessing
import os
import pathlib
import shutil
import tempfile
import time
import uuid
from concurrent.futures import Future, ProcessPoolExecutor
from enum import Enum
from typing import Any, Optional

from composer.utils.dist import broadcast_object_list, get_global_rank, get_local_rank
from composer.utils.file_helpers import (
    parse_uri,
    validate_credentials,
)
from composer.utils.object_store.mlflow_object_store import MLFLOW_DBFS_PATH_PREFIX, MLFlowObjectStore
from composer.utils.object_store.object_store import (
    ObjectStore,
    ObjectStoreTransientError,
)
from composer.utils.object_store.utils import build_remote_backend
from composer.utils.retrying import retry

log = logging.getLogger(__name__)

__all__ = ['RemoteUploader']


class RemoteFilesExistingCheckStatus(Enum):
    EXIST = 1
    TIMEOUT = 2
    ERROR = 3


def _check_remote_files_exists(
    remote_backend_name: str,
    backend_kwargs: dict[str, Any],
    remote_checkpoint_file_names: list[str],
    main_process_pid: int,
    is_remote_upload_failed: multiprocessing.Event, # pyright: ignore[reportGeneralTypeIssues]
    max_wait_time_in_seconds: int = 3600,
    wait_before_next_try_in_seconds: float = 30,
):
    start_time = time.time()
    object_store = build_remote_backend(remote_backend_name, backend_kwargs)

    for remote_file_name in remote_checkpoint_file_names:
        while True:
            if is_remote_upload_failed.is_set():
                log.debug(f'Stop symlink uploading since the checkpoint files uploading failed')
                return RemoteFilesExistingCheckStatus.ERROR
            # Return if parent process exits
            try:
                os.kill(main_process_pid, 0)
            except OSError:
                return RemoteFilesExistingCheckStatus.ERROR
            try:
                object_store.get_object_size(remote_file_name)
                break
            except Exception as e:
                if not isinstance(e, FileNotFoundError):
                    log.debug(f'Got exception {type(e)}: {str(e)} when accessing remote file {remote_file_name}')
                time.sleep(wait_before_next_try_in_seconds)
            if time.time() - start_time > max_wait_time_in_seconds:
                return RemoteFilesExistingCheckStatus.TIMEOUT
    return RemoteFilesExistingCheckStatus.EXIST


def _upload_file_to_object_store(
    remote_backend_name: str,
    backend_kwargs: dict[str, Any],
    remote_file_name: str,
    local_file_path: str,
    overwrite: bool,
    num_attempts: int,
) -> int:
    object_store = build_remote_backend(remote_backend_name, backend_kwargs)

    @retry(ObjectStoreTransientError, num_attempts=num_attempts)
    def upload_file(retry_index: int = 0):
        if retry_index == 0 and not overwrite:
            try:
                object_store.get_object_size(remote_file_name)
            except FileNotFoundError:
                # Good! It shouldn't exist.
                pass
            else:
                raise FileExistsError(
                    f'Object {remote_file_name} already exists, but overwrite was set to False. '
                    'Please set `save_overwrite` to `True` in Trainer to overwrite the existing file.',
                )
        log.info(f'Uploading file {local_file_path} to {remote_file_name}')
        object_store.upload_object(
            object_name=remote_file_name,
            filename=local_file_path,
        )
        os.remove(local_file_path)

    log.info(f'Finished uploading file {local_file_path} to {remote_file_name}')
    # When encountering issues with too much concurrency in uploads, staggering the uploads can help.
    # This stagger is intended for use when uploading model shards from every rank, and will effectively reduce
    # the concurrency by a factor of num GPUs per node.
    local_rank = get_local_rank()
    local_rank_stagger = int(os.environ.get('COMPOSER_LOCAL_RANK_STAGGER_SECONDS', 0))
    log.debug(f'Staggering uploads by {local_rank * local_rank_stagger} seconds on {local_rank} local rank.')
    time.sleep(local_rank * local_rank_stagger)
    upload_file()
    return 0


class RemoteUploader:
    """Class for uploading a file to object store asynchronously."""

    def __init__(
        self,
        remote_folder: str,
        backend_kwargs: Optional[dict[str, Any]] = None,
        num_concurrent_uploads: int = 2,
        num_attempts: int = 3,
    ):
        if num_concurrent_uploads < 1 or num_attempts < 1:
            raise ValueError(
                f'num_concurrent_uploads and num_attempts must be >= 1, but got {num_concurrent_uploads} and {num_attempts}',
            )

        self.remote_folder = remote_folder
        # A folder to use for staging uploads
        self._tempdir = tempfile.TemporaryDirectory()
        self._upload_staging_folder = self._tempdir.name
        self.remote_backend_name, self.remote_bucket_name, self.path = parse_uri(remote_folder)

        self.backend_kwargs: dict[str, Any] = backend_kwargs if backend_kwargs is not None else {}
        if self.remote_backend_name in ['s3', 'oci', 'gs'] and 'bucket' not in self.backend_kwargs:
            self.backend_kwargs['bucket'] = self.remote_bucket_name
        elif self.remote_backend_name == 'libcloud':
            if 'container' not in self.backend_kwargs:
                self.backend_kwargs['container'] = self.remote_bucket_name
        elif self.remote_backend_name == 'azure':
            self.remote_backend_name = 'libcloud'
            self.backend_kwargs = {
                'provider': 'AZURE_BLOBS',
                'container': self.remote_bucket_name,
                'key_environ': 'AZURE_ACCOUNT_NAME',
                'secret_environ': 'AZURE_ACCOUNT_ACCESS_KEY',
            }
        elif self.remote_backend_name == 'dbfs':
            self.backend_kwargs['path'] = self.path
        elif self.remote_backend_name == 'wandb':
            raise NotImplementedError(
                f'There is no implementation for WandB via URI. Please use '
                'WandBLogger with log_artifacts set to True.',
            )
        else:
            raise NotImplementedError(
                f'There is no implementation for the cloud backend {self.remote_backend_name} via URI. Please use '
                'one of the supported object stores (s3, oci, gs, azure, dbfs).',
            )

        self.num_attempts = num_attempts
        self._remote_backend: Optional[ObjectStore] = None
        mp_context = multiprocessing.get_context('spawn')
        self.upload_executor = ProcessPoolExecutor(
            max_workers=num_concurrent_uploads,
            mp_context=mp_context,
        )
        self.check_remote_files_exist_executor = ProcessPoolExecutor(
            max_workers=2,
            mp_context=mp_context,
        )
        self.is_remote_upload_failed = mp_context.Manager().Event()

        # Used internally to track the future status.
        # If a future completed successfully, we'll remove it from this list
        # when check_workers() or wait() is called
        self.futures: list[Future] = []

        self.pid = os.getpid()

    @property
    def remote_backend(self) -> ObjectStore:
        if self._remote_backend is None:
            self._remote_backend = build_remote_backend(self.remote_backend_name, self.backend_kwargs)
        return self._remote_backend

    def init(self):
        # If it's dbfs path like: dbfs:/databricks/mlflow-tracking/{mlflow_experiment_id}/{mlflow_run_id}/
        # We need to fill out the experiment_id and run_id

        if get_global_rank() == 0:

            @retry(ObjectStoreTransientError, num_attempts=self.num_attempts)
            def _validate_credential_with_retry():
                validate_credentials(self.remote_backend, '.credentials_validated_successfully')

            _validate_credential_with_retry()
        if self.path.startswith(MLFLOW_DBFS_PATH_PREFIX):
            if get_global_rank() == 0:
                assert isinstance(self.remote_backend, MLFlowObjectStore)
                self.path = self.remote_backend.get_dbfs_path(self.path)
            path_list = [self.path]
            broadcast_object_list(path_list, src=0)
            self.path = path_list[0]
            self.backend_kwargs['path'] = self.path

    def upload_file_async(
        self,
        remote_file_name: str,
        file_path: pathlib.Path,
        overwrite: bool,
    ):
        """Async call to submit a job for uploading.

        It returns a future, so users can track the status of the individual future.
        User can also call wait() to wait for all the futures.
        """
        # Copy file to staging folder
        copied_path = os.path.join(self._upload_staging_folder, str(uuid.uuid4()))
        os.makedirs(self._upload_staging_folder, exist_ok=True)
        shutil.copy2(file_path, copied_path)

        # Async upload file
        future = self.upload_executor.submit(
            _upload_file_to_object_store,
            remote_backend_name=self.remote_backend_name,
            backend_kwargs=self.backend_kwargs,
            remote_file_name=remote_file_name,
            local_file_path=copied_path,
            overwrite=overwrite,
            num_attempts=self.num_attempts,
        )
        self.futures.append(future)
        return future

    def check_workers(self):
        """Non-blocking call to check workers are either running or done.

        Traverse self.futures, and check if it's completed
        1. if it completed with exception, raise that exception
        2. if it completed without exception, remove it from self.futures
        """
        done_futures: list[Future] = []
        for future in self.futures:
            if future.done():
                # future.exception is a blocking call
                exception_or_none = future.exception()
                if exception_or_none is not None:
                    self.is_remote_upload_failed.set()
                    raise exception_or_none
                else:
                    done_futures.append(future)
        for future in done_futures:
            self.futures.remove(future)

    def wait(self):
        """Blocking call to wait all the futures to complete.

        If a future is done successfully, remove it from self.futures(),
        otherwise, raise the exception
        """
        for future in self.futures:
            exception_or_none = future.exception()
            if exception_or_none is not None:
                self.is_remote_upload_failed.set()
                raise exception_or_none
        self.futures = []

    def wait_and_close(self):
        """Blocking call to wait all uploading to finish and close this uploader.

        After this function is called, users can not use this uploader
        to uploading anymore. So please only call wait_and_close() after submitting
        all uploading requests.
        """
        # make sure all workers are either running, or completed successfully
        self.wait()
        self.upload_executor.shutdown(wait=True)
        self.check_remote_files_exist_executor.shutdown(wait=True)
        log.debug('Finished all uploading tasks, closing RemoteUploader')

    def check_remote_files_exist_async(
        self,
        remote_checkpoint_file_names: list[str],
        max_wait_time_in_seconds: int = 3600,
        wait_before_next_try_in_seconds: float = 30,
    ):
        future = self.check_remote_files_exist_executor.submit(
            _check_remote_files_exists,
            remote_backend_name=self.remote_backend_name,
            backend_kwargs=self.backend_kwargs,
            remote_checkpoint_file_names=remote_checkpoint_file_names,
            main_process_pid=self.pid,
            is_remote_upload_failed=self.is_remote_upload_failed,
            max_wait_time_in_seconds=max_wait_time_in_seconds,
            wait_before_next_try_in_seconds=wait_before_next_try_in_seconds,
        )
        self.futures.append(future)
        return future



# Copyright 2024 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""Helpers for working with object stores."""

from typing import Any

from composer.utils.object_store.gcs_object_store import GCSObjectStore
from composer.utils.object_store.libcloud_object_store import LibcloudObjectStore
from composer.utils.object_store.mlflow_object_store import MLFLOW_DBFS_PATH_PREFIX, MLFlowObjectStore
from composer.utils.object_store.oci_object_store import OCIObjectStore
from composer.utils.object_store.s3_object_store import S3ObjectStore
from composer.utils.object_store.sftp_object_store import SFTPObjectStore
from composer.utils.object_store.uc_object_store import UCObjectStore

__all__ = ['build_remote_backend']


def build_remote_backend(remote_backend_name: str, backend_kwargs: dict[str, Any]):
    """Build object store given the backend name and kwargs."""
    remote_backend_cls = None
    remote_backend_name_to_cls = {
        's3': S3ObjectStore,
        'oci': OCIObjectStore,
        'sftp': SFTPObjectStore,
        'libcloud': LibcloudObjectStore,
        'gs': GCSObjectStore,
    }

    # Handle `dbfs` backend as a special case, since it can map to either :class:`.UCObjectStore`
    # or :class:`.MLFlowObjectStore`.
    if remote_backend_name == 'dbfs':
        path = backend_kwargs['path']
        if path.startswith(MLFLOW_DBFS_PATH_PREFIX):
            remote_backend_cls = MLFlowObjectStore
        else:
            # Validate if the path conforms to the requirements for UC volume paths
            UCObjectStore.validate_path(path)
            remote_backend_cls = UCObjectStore
    else:
        remote_backend_cls = remote_backend_name_to_cls.get(remote_backend_name, None)
        if remote_backend_cls is None:
            supported_remote_backends = list(remote_backend_name_to_cls.keys()) + ['dbfs']
            raise ValueError(
                f'The remote backend {remote_backend_name} is not supported. Please use one of ({supported_remote_backends})',
            )

    return remote_backend_cls(**backend_kwargs)



# Copyright 2024 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""API for loading checkpoints."""

import contextlib
import logging
import os
import pickle
import tempfile
import textwrap
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Optional, Sequence, Union

import torch
import torch.distributed.checkpoint as DCP
from packaging import version
from torch import nn
from torch.distributed.fsdp import FullyShardedDataParallel as FSDP
from torch.optim import Optimizer

from composer.checkpoint.download import download_and_extract_symlink, download_monolithic_checkpoint
from composer.checkpoint.state_dict import (
    _cast_state_dict_to_precision,
    _extract_keys_from_state_dict,
    _is_model_fsdp,
    _remove_keys_from_state_dict,
    get_model_state_dict,
    get_optim_state_dict,
)
from composer.core.state import State
from composer.distributed import prepare_fsdp_module
from composer.models import ComposerModel
from composer.utils import FSDPConfig, dist, reproducibility
from composer.utils.checkpoint import (
    DistCPObjectStoreReader,
    FileSystemReaderWithValidation,
    _ensure_valid_checkpoint,
    _torch_load_with_validation,
)
from composer.utils.file_helpers import is_uri

log = logging.getLogger(__name__)


@dataclass
class CheckpointLoadOptions:
    """Options for loading a checkpoint.

    Args:
        load_model (bool): Whether to load the model checkpoint.
        load_optimizer (bool): Whether to load the optimizer checkpoint.
        load_resumption_state (bool): Whether to load the resumption state.
        include_keys (Optional[Union[str, Sequence[str]]]): Specific keys to include in the state dict.
        ignore_keys (Optional[Union[str, Sequence[str]]]): Specific keys to ignore in the state dict.
        sharded_checkpoint (bool): Whether the checkpoint is sharded or not.
        shard_as_needed_during_load (bool): Whether to shard the model and optimizer during load. If model is unsharded and checkpoint is sharded, model will be sharded before.
            If model is unsharded and checkpoint is unsharded, model will be sharded after loading the checkpoint.
        strict (bool): Whether to load the checkpoint strictly.
        precision (str): The precision to cast the model/optim to after loading
        cpu_offload (bool): Whether to offload the state dict to CPU before loading.
        load_planner (Optional[LoadPlanner]): The load planner to use for loading the checkpoint.
        fsdp_config (Optional[Union[FSDPConfig, dict]]): The FSDP config to use for sharding the model and optimizer.
        device_mesh (Optional[DeviceMesh]): The device mesh to use for loading the checkpoint.
        seed (Optional[int]): The seed to use for sharding the model and optimizer.
    """
    load_model: bool = True
    load_optimizer: bool = False
    load_resumption_state: bool = False  # dataset_state, timestamp,
    # Specific key-level loading configs
    # e.g.'model.layers.transformer.0.w1.weight'
    include_keys: Optional[Union[str, Sequence[str]]] = None
    ignore_keys: Optional[Union[str, Sequence[str]]] = None
    sharded_checkpoint: bool = False  # TODO: Auto-detect sharded
    shard_as_needed_during_load: bool = False
    strict: bool = True
    precision: str = 'fp32'
    cpu_offload: bool = True
    load_planner: Optional[Any] = None
    fsdp_config: Optional[Union[FSDPConfig, dict]] = None
    device_mesh: Optional[Any] = None
    seed: Optional[int] = 42


def load_checkpoint(
    load_path: str,
    load_options: Optional[Union[CheckpointLoadOptions, dict]],
    model: Optional[Union[ComposerModel, nn.Module]] = None,
    optim: Optional[Optimizer] = None,
    state: Optional[State] = None,
    model_child_path: Optional[str] = None,
    optim_child_path: Optional[str] = None,
    resumption_filename: str = 'resumption.pkl',
):
    """Optionally download and load a checkpoint according to the options into specified state.

    Args:
        load_path (str): The path or uri to the checkpoint to load or symlink to the path/uri. If URI specified then files will be downloaded first.
        load_options (CheckpointLoadOptions): The options for loading the checkpoint.
        model (Optional[ComposerModel]): The model to load the checkpoint into. If not provided, the model from the state will be used.
        optim (Optional[Optimizer]): The optimizer to load the checkpoint into. If not provided, the optimizer from the state will be used.
        state (Optional[State]): The state to load the checkpoint. If not provided, the model and optimizer must be provided.
        model_child_path (Optional[str]): The path to the model checkpoint within the load_path.
            E.g. if your checkpoints are unsharded and saved in load_path/my_cool_model/my_model.pt, set model_child_path='my_cool_model/my_model.pt'. If not specified, 'model/model.pt' is used.
            if your checkpoints are sharded and saved in load_path/my_sharded_model/ then set model_child_path='my_sharded_model'. If not specified, 'model' is used.
        optim_child_path (Optional[str]): The path to the optimizer checkpoint within the load_path.
            e.g. if your checkpoints are unsharded and saved in load_path/my_cool_optim/my_optim.pt, set optim_child_path='my_cool_optim/my_optim.pt'. If not specified, 'optim/optim.pt' is used.
            if your checkpoints are sharded and saved in load_path/my_sharded_optim/ then set optim_child_path='my_sharded_optim'. If not specified, 'optim' is used.
        resumption_filename (Optional[str]): The filename of the resumption state file. Default is 'resumption.pkl'.
    """
    if load_options is None:
        load_options = CheckpointLoadOptions()
    if isinstance(load_options, dict):
        load_options = CheckpointLoadOptions(**load_options)

    if load_options.load_model:
        if model is None:
            if state is None:
                raise ValueError('Model or state must be provided if loading model checkpoint.')
            model = state.model
        if model_child_path is None:
            model_child_path = 'model' if load_options.sharded_checkpoint else 'model/model.pt'

    if load_options.load_optimizer:
        if optim is None:
            if state is None:
                raise ValueError('Optimizer or state must be provided if loading optimizer checkpoint.')
            optim = state.optimizers[0]
        if optim_child_path is None:
            optim_child_path = 'optim' if load_options.sharded_checkpoint else 'optim/optim.pt'

    if load_options.load_resumption_state:
        if state is None:
            raise ValueError('State must be provided if loading resumption state.')
        load_resumption_checkpoint(state, os.path.join(load_path, resumption_filename))

    if load_options.load_model:
        assert model is not None
        assert model_child_path is not None
        model_load_path = os.path.join(load_path, model_child_path)
        if state is not None:
            state.automicrobatch_fsdp_hook_handles, state.fsdp_modules = load_model_checkpoint(
                model,
                load_path=model_load_path,
                load_options=load_options,
            )
        else:
            load_model_checkpoint(
                model,
                load_path=model_load_path,
                load_options=load_options,
            )

    if load_options.load_optimizer:
        assert optim_child_path is not None
        optim_load_path = os.path.join(load_path, optim_child_path)
        assert model is not None
        assert optim is not None
        load_optim_checkpoint(
            model,
            optim,
            optim_load_path,
            load_options=load_options,
        )


def load_model_checkpoint(
    model: Union[ComposerModel, nn.Module],
    load_path: Optional[str] = None,
    load_options: Optional[Union[CheckpointLoadOptions, dict]] = None,
    seed: int = 42,
) -> tuple[list, dict]:
    """Load a a model checkpoint from the specified path into the model.

    Args:
        model (Union[ComposerModel, nn.Module]): The model to load the checkpoint into.
        load_path (Optional[str]): The path to the checkpoint to load.
        load_options (Optional[Union[CheckpointLoadOptions, Dict]]): The options for loading the checkpoint.
        seed (int): The seed to use for sharding the model and optimizer.
    """
    if load_options is None:
        load_options = CheckpointLoadOptions()
    elif isinstance(load_options, dict):
        load_options = CheckpointLoadOptions(**load_options)
    if load_options.include_keys is not None and load_options.ignore_keys is not None:
        raise ValueError('Only one of include_keys or ignore_keys can be specified.')

    if load_options.include_keys is not None or load_options.ignore_keys is not None:
        load_options.strict = False

    automicrobatch_fsdp_hook_handles = []
    fsdp_modules = {}

    if load_options.sharded_checkpoint:
        if not _is_model_fsdp(model):
            if load_options.shard_as_needed_during_load:
                automicrobatch_fsdp_hook_handles, fsdp_modules = _shard_with_fsdp(
                    model,
                    fsdp_config=load_options.fsdp_config,
                    precision=load_options.precision,
                    seed=seed,
                )
            else:
                raise ValueError(
                    'Model is not sharded but checkpoint is sharded. Please pass in a model wrapped with FSDP or set shard_model_as_needed_during_load to True.',
                )
        assert load_path is not None
        _load_sharded_model_checkpoint(model, load_path=load_path, load_options=load_options)
    else:
        if dist.get_global_rank() == 0:
            assert load_path is not None
            _load_unsharded_model_checkpoint(model, load_path=load_path, load_options=load_options)

        if load_options.shard_as_needed_during_load and not _is_model_fsdp(model):
            if load_options.fsdp_config is None:
                load_options.fsdp_config = {'sync_module_states': True}
            elif isinstance(load_options.fsdp_config, dict):
                load_options.fsdp_config.update({'sync_module_states': True})
            else:
                load_options.fsdp_config.sync_module_states = True
            automicrobatch_fsdp_hook_handles, fsdp_modules = _shard_with_fsdp(
                model,
                fsdp_config=load_options.fsdp_config,
                precision=load_options.precision,
                seed=seed,
            )
    return automicrobatch_fsdp_hook_handles, fsdp_modules


def _shard_with_fsdp(
    model: Union[ComposerModel, nn.Module],
    optimizer: Optional[Optimizer] = None,
    fsdp_config: Optional[Union[FSDPConfig, dict]] = None,
    precision: Optional[str] = None,
    seed: int = 42,
) -> tuple[list, dict]:
    if fsdp_config is None:
        fsdp_config = FSDPConfig()
    if isinstance(fsdp_config, dict):
        fsdp_config = FSDPConfig(**fsdp_config)
    with reproducibility.seed_context(seed):
        automicrobatch_fsdp_hook_handles, fsdp_modules = prepare_fsdp_module(
            model,
            optimizers=optimizer,
            fsdp_config=fsdp_config,
            precision=precision,
        )
    return automicrobatch_fsdp_hook_handles, fsdp_modules


def _load_sharded_model_checkpoint(
    model: Union[ComposerModel, nn.Module],
    load_path: str,
    load_options: Optional[Union[CheckpointLoadOptions, dict]],
):
    if load_options is None:
        load_options = CheckpointLoadOptions()
    elif isinstance(load_options, dict):
        load_options = CheckpointLoadOptions(**load_options)
    model_state_dict = get_model_state_dict(
        model,
        sharded_state_dict=True,
        include_keys=load_options.include_keys,
        ignore_keys=load_options.ignore_keys,
    )
    model_state_dict = download_and_load_sharded_state_dict(
        load_path,
        load_options.device_mesh,
        model_state_dict,
        load_options.load_planner,
    )
    model_state_dict = _cast_state_dict_to_precision(model_state_dict, load_options.precision)
    # TODO: raise warning for unknown or missing keys.
    log.debug(f'Loading sharded state dict into model.')
    if version.parse(torch.__version__) >= version.parse('2.3.0') and dist.is_initialized():
        torch_set_model_state_dict(
            model,
            model_state_dict,
            strict=load_options.strict,
            cpu_offload=load_options.cpu_offload,
            sharded_state_dict=True,
        )
    else:
        _load_model_state_dict_with_fsdp_context_manager(
            model,
            model_state_dict,
            sharded_state_dict=True,
            strict=load_options.strict,
        )


def _load_unsharded_model_checkpoint(
    model: Union[ComposerModel, nn.Module],
    load_path: str,
    load_options: Optional[Union[CheckpointLoadOptions, dict]] = None,
):
    if dist.get_global_rank() != 0:
        return
    if load_options is None:
        load_options = CheckpointLoadOptions()
    elif isinstance(load_options, dict):
        load_options = CheckpointLoadOptions(**load_options)
    if _is_model_fsdp(model):
        raise ValueError(
            'Model is sharded but checkpoint is not sharded. Please pass in a model not wrapped with FSDP.',
        )
    load_path_is_remote = is_uri(load_path)
    download_dir_context = tempfile.TemporaryDirectory if load_path_is_remote else contextlib.nullcontext
    with download_dir_context() as download_dir:
        file_path = load_path
        if load_path_is_remote:
            filename = Path(load_path).name
            assert download_dir is not None
            file_path = os.path.join(download_dir, filename)
            download_monolithic_checkpoint(load_path, download_dir)

        model_state_dict = _torch_load_with_validation(file_path, map_location='cpu')
        model_state_dict = _cast_state_dict_to_precision(model_state_dict, load_options.precision)
        if load_options.include_keys is not None:
            model_state_dict = _extract_keys_from_state_dict(model_state_dict, load_options.include_keys)
        if load_options.ignore_keys is not None:
            model_state_dict = _remove_keys_from_state_dict(model_state_dict, load_options.ignore_keys)
        model.load_state_dict(model_state_dict, strict=load_options.strict)


def load_optim_checkpoint(
    model: Union[ComposerModel, nn.Module],
    optim: torch.optim.Optimizer,
    load_path: str,
    load_options: Optional[Union[CheckpointLoadOptions, dict]] = None,
):
    """Load an optimizer checkpoint from the specified path into the optimizer.

    Args:
        model (Union[ComposerModel, nn.Module]): The model to load the optimizer checkpoint into.
        optim (torch.optim.Optimizer): The optimizer to load the checkpoint into.
        load_path (Optional[str]): The path to the checkpoint to load.
        load_options (Optional[Union[CheckpointLoadOptions, Dict]]): The options for loading the checkpoint.
    """
    if load_options is None:
        load_options = CheckpointLoadOptions()
    elif isinstance(load_options, dict):
        load_options = CheckpointLoadOptions(**load_options)
    if load_options.sharded_checkpoint and not _is_model_fsdp(model):
        raise ValueError(
            textwrap.dedent(
                """Model/Optim is not sharded but checkpoint is sharded.
                                         Please  pass in a model/optim wrapped with FSDP.""",
            ),
        )
    if not load_options.sharded_checkpoint and not _is_model_fsdp(model) and load_options.shard_as_needed_during_load:
        raise ValueError(
            textwrap.dedent(
                """Neither model nor optim nor checkpoint is sharded, but shard_as_needed_during_load is set.
                                         Sharding the optim after load is not supported. please set shard_as_needed_during_load to False.
                                         """,
            ),
        )

    if load_options.shard_as_needed_during_load and _is_model_fsdp(model):
        raise NotImplementedError(
            'Loading optimizers for models that have been sharded during load (either before or after loading in the model checkpoint) is not currently supported.',
        )

    if load_options.sharded_checkpoint:
        _load_sharded_optim_checkpoint(model=model, optim=optim, load_path=load_path, load_options=load_options)
    else:
        _load_unsharded_optim_checkpoint(
            model=model,
            optim=optim,
            load_path=load_path,
            precision=load_options.precision,
        )


def _load_sharded_optim_checkpoint(
    model: Union[ComposerModel, nn.Module],
    optim: torch.optim.Optimizer,
    load_path: str,
    load_options: CheckpointLoadOptions,
):
    if not _is_model_fsdp(model):
        raise ValueError(
            'Model is not sharded but checkpoint is sharded. Please either use load_model_checkpoint(model, load_path, optimizer, shard_as_needed_during_load=True) or pass in a model wrapped with FSDP.',
        )
    # if not _is_optimizer_sharded(optim):
    #     raise ValueError("Optimizer is not sharded but checkpoint is sharded. Please pass in a sharded optimizer by passing a sharded model's parameters to an optimizer constructor.")
    optim_state_dict = get_optim_state_dict(model, optim, sharded_state_dict=True)
    optim_state_dict = download_and_load_sharded_state_dict(
        load_path=load_path,
        device_mesh=load_options.device_mesh,
        state_dict=optim_state_dict,
        load_planner=load_options.load_planner,
    )
    for param_key, param_state_dict in optim_state_dict['state'].items():
        optim_state_dict['state'][param_key] = _cast_state_dict_to_precision(param_state_dict, load_options.precision)
    if version.parse(torch.__version__) >= version.parse('2.3.0') and dist.is_initialized():
        torch_set_optimizer_state_dict(
            model,
            optim,
            optim_state_dict,
            strict=load_options.strict,
            cpu_offload=load_options.cpu_offload,
            sharded_state_dict=True,
        )
    else:
        _load_optim_state_dict_with_fsdp_context_manager(
            model,
            optim,
            optim_state_dict,
            sharded_state_dict=True,
            strict=load_options.strict,
        )


def _load_unsharded_optim_checkpoint(
    model: Union[ComposerModel, nn.Module],
    optim: torch.optim.Optimizer,
    load_path: str,
    precision: str = 'fp32',
):
    if dist.get_global_rank() != 0:
        return
    if _is_model_fsdp(model):
        raise ValueError('Model is sharded, but checkpoint is not sharded. Please pass in a model unwrapped from FSDP.')
    # if _is_optimizer_sharded(optim):
    #     raise ValueError("Optimizer is sharded, but checkpoint is not sharded. Please pass in an unsharded optimizer by passing an unsharded model's parameters to an optimizer constructor.")
    load_path_is_remote = is_uri(load_path)
    download_dir_context = tempfile.TemporaryDirectory if load_path_is_remote else contextlib.nullcontext
    with download_dir_context() as download_dir:
        file_path = load_path
        if load_path_is_remote:
            filename = Path(load_path).name
            assert download_dir is not None
            file_path = os.path.join(download_dir, filename)
            download_monolithic_checkpoint(load_path, file_path)

        optim_state_dict = _torch_load_with_validation(file_path, map_location='cpu')
        for param_key, param_state_dict in optim_state_dict['state'].items():
            optim_state_dict['state'][param_key] = _cast_state_dict_to_precision(param_state_dict, precision)
        optim.load_state_dict(optim_state_dict)


def _preprocess_local_load_path(load_path: str):
    if os.path.islink(load_path):
        load_path = os.path.join(os.path.dirname(load_path), os.readlink(load_path))
    if os.path.exists(load_path):
        if not os.path.isdir(load_path):
            raise ValueError(f'load_path must be a directory when using sharded state dict. Got {load_path}')
    else:
        raise FileNotFoundError(f'{load_path} not found!')
    return load_path


def _load_model_state_dict_with_fsdp_context_manager(
    model: nn.Module,
    model_state_dict: dict,
    sharded_state_dict: bool,
    strict: bool,
):
    from torch.distributed.fsdp.fully_sharded_data_parallel import (
        FullStateDictConfig,
        ShardedStateDictConfig,
        StateDictType,
    )
    state_dict_type = StateDictType.SHARDED_STATE_DICT if sharded_state_dict else StateDictType.FULL_STATE_DICT
    state_dict_config = ShardedStateDictConfig(offload_to_cpu=True,) if sharded_state_dict else FullStateDictConfig(
        offload_to_cpu=True,
        rank0_only=True,
    )
    with FSDP.state_dict_type(model, state_dict_type=state_dict_type, state_dict_config=state_dict_config):
        missing_keys, unexpected_keys = model.load_state_dict(
            model_state_dict,
            strict=strict,
        )

    if len(missing_keys) > 0:
        log.warning(f"Found these missing keys in the checkpoint: {', '.join(missing_keys)}")
    if len(unexpected_keys) > 0:
        log.warning(f"Found these unexpected keys in the checkpoint: {', '.join(unexpected_keys)}")


def _load_optim_state_dict_with_fsdp_context_manager(
    model: nn.Module,
    optim,
    optim_state_dict: dict,
    sharded_state_dict: bool,
    strict: bool,
):
    from torch.distributed.fsdp.fully_sharded_data_parallel import (
        FullOptimStateDictConfig,
        ShardedOptimStateDictConfig,
        StateDictType,
    )
    state_dict_type = StateDictType.SHARDED_STATE_DICT if sharded_state_dict else StateDictType.FULL_STATE_DICT
    optim_state_dict_config = optim_state_dict_config = ShardedOptimStateDictConfig(
    ) if sharded_state_dict else FullOptimStateDictConfig(offload_to_cpu=True, rank0_only=True)
    with FSDP.state_dict_type(model, state_dict_type=state_dict_type, optim_state_dict_config=optim_state_dict_config):
        optim_state_dict = FSDP.optim_state_dict_to_load(  #  type: ignore
                                optim_state_dict=optim_state_dict, model=model, optim=optim,
                            )
    assert optim_state_dict is not None
    optim.load_state_dict(optim_state_dict)


def torch_set_model_state_dict(
    model: torch.nn.Module,
    model_state_dict: dict,
    strict: bool,
    cpu_offload: bool,
    sharded_state_dict: bool = True,
):
    """Set the model state dict with the given options.

    Args:
        model (torch.nn.Module): The model to set the state dict to.
        model_state_dict (dict): The model state dict to set.
        strict (bool): Whether to set the state dict strictly.
        cpu_offload (bool): Whether to offload the state dict to CPU before setting.
        sharded_state_dict (bool): Whether the state dict is sharded or not.
    """
    if version.parse(torch.__version__) >= version.parse('2.3.0') and dist.is_initialized():
        from torch.distributed.checkpoint.state_dict import StateDictOptions, set_model_state_dict
        try:
            set_model_state_dict(
                model,
                model_state_dict,
                options=StateDictOptions(
                    strict=strict,
                    cpu_offload=cpu_offload,
                    full_state_dict=not sharded_state_dict,
                ),
            )
        except AttributeError as e:
            # Issue: https://github.com/pytorch/pytorch/issues/127351
            if "ShardedTensor' object has no attribute 'placements'" in str(e):
                raise RuntimeError(
                    textwrap.dedent(
                        'PyTorch DTensor broke backwards compatibility in older checkpoints '
                        'with ShardedTensor, which is now deprecated. To load old checkpoints, '
                        'either downgrade to PyTorch <2.3.0 or explicitly pass process groups '
                        'in the Trainer constructor via '
                        "`parallelism_config = {'fsdp': {'process_group': 'mod1'}}`. We can "
                        'provide assistance at https://github.com/mosaicml/composer/issues.',
                    ),
                ) from e
            else:
                raise e


def torch_set_optimizer_state_dict(
    model: torch.nn.Module,
    optim: torch.optim.Optimizer,
    optim_state_dict: dict,
    strict: bool,
    cpu_offload: bool,
    sharded_state_dict: bool = True,
):
    """Set the optimizer state dict with the given options.

    Args:
        model (torch.nn.Module): The model to set the state dict to.
        optim (torch.optim.Optimizer): The optimizer to set the state dict to.
        optim_state_dict (dict): The optimizer state dict to set.
        strict (bool): Whether to set the state dict strictly.
        cpu_offload (bool): Whether to offload the state dict to CPU before setting.
        sharded_state_dict (bool): Whether the state dict is sharded or not.
    """
    if version.parse(torch.__version__) >= version.parse('2.3.0') and dist.is_initialized():
        from torch.distributed.checkpoint.state_dict import StateDictOptions, set_optimizer_state_dict
        set_optimizer_state_dict(
            model=model,
            optimizers=optim,
            optim_state_dict=optim_state_dict,
            options=StateDictOptions(
                full_state_dict=not sharded_state_dict,
                strict=strict,
                cpu_offload=cpu_offload,
            ),
        )


def download_and_load_sharded_state_dict(
    load_path: str,
    device_mesh: Optional[Any],
    state_dict: dict,
    load_planner: Optional[Any] = None,
):
    """A helper function to download and load a sharded state dict.

    Args:
        load_path (str): The path to the checkpoint to load.
        device_mesh (Optional[Any]): The device mesh to use for loading the checkpoint.
        state_dict (dict): The state dict to load.
        load_planner (Optional[Any]): The load planner to use for loading the checkpoint.
    """
    load_path_is_remote = is_uri(load_path)
    download_dir_context = tempfile.TemporaryDirectory if load_path_is_remote else contextlib.nullcontext
    with download_dir_context() as download_dir:
        if load_path_is_remote:
            local_rank0_index = dist.get_global_rank() - dist.get_local_rank()
            rank0_download_tempdir = str(dist.all_gather_object(download_dir)[local_rank0_index])
            if load_path.endswith('.symlink'):
                load_path = download_and_extract_symlink(load_path)
            storage_reader = DistCPObjectStoreReader(
                source_path=load_path,
                destination_path=str(Path(rank0_download_tempdir) / Path('checkpoints')),
                device_mesh=device_mesh,
            )
        else:
            load_path = _preprocess_local_load_path(load_path)
            storage_reader = FileSystemReaderWithValidation(load_path)

            try:
                storage_reader.read_metadata()
            except AttributeError as e:
                if '_MEM_FORMAT_ENCODING' in str(e):
                    raise ValueError(
                        'Unable to read checkpoint metadata. The checkpoint was likely saved with a '
                        'newer version of torch. Upgrade your torch version to load this checkpoint.',
                    )
                else:
                    raise

        log.debug(f'Loading sharded state dict from {load_path} into memory.')
        if version.parse(torch.__version__) < version.parse('2.2.0'):
            DCP.load_state_dict(state_dict=state_dict, storage_reader=storage_reader, planner=load_planner)
        else:
            DCP.load(state_dict=state_dict, storage_reader=storage_reader, planner=load_planner)
    return state_dict


def load_resumption_checkpoint(state: State, load_path: str):
    """Load the resumption state from the specified path into the state.

    Args:
        state (State): The state to load the resumption state into.
        load_path (str): The path to the resumption state to load.

    """
    load_path = str(_ensure_valid_checkpoint(load_path))
    resumption_state_dict = pickle.load(open(load_path, 'rb'))
    if 'dataset_state' in resumption_state_dict:
        state._load_dataset_state(resumption_state_dict['dataset_state'])
    if 'timestamp' in resumption_state_dict:
        state.timestamp.load_state_dict(resumption_state_dict['timestamp'])

    if 'schedulers' in resumption_state_dict:
        for scheduler in state.schedulers:
            fqn = type(scheduler).__qualname__
            scheduler.load_state_dict(resumption_state_dict['schedulers'][fqn])

    if 'algorithms' in resumption_state_dict:
        for algorithm in state.algorithms:
            fqn = type(algorithm).__qualname__
            algorithm.load_state_dict(dict(resumption_state_dict['algorithms'])[fqn])

    if 'callbacks' in resumption_state_dict:
        for callback in state.callbacks:
            fqn = type(callback).__qualname__
            callback.load_state_dict(dict(resumption_state_dict['callbacks'])[fqn])

    if 'scaler' in resumption_state_dict:
        fqn = type(state.scaler).__qualname__
        assert state.scaler is not None
        state.scaler.load_state_dict(resumption_state_dict['scaler'][fqn])

    state.rank_zero_seed = resumption_state_dict['rank_zero_seed']
    return resumption_state_dict['rng']



# Copyright 2024 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""Load a checkpoint."""
import logging
from typing import Optional, Union

from composer.core import Callback, State
from composer.core.event import Event
from composer.loggers import Logger
from composer.models.huggingface import HuggingFaceModel
from composer.utils.checkpoint import load_checkpoint
from composer.utils.file_helpers import maybe_create_object_store_from_uri, parse_uri

log = logging.getLogger(__name__)


class LoadCheckpoint(Callback):
    """Callback that loads a checkpoint at the specified event.

    Args:
        load_path (str): The path to the checkpoint to load.
        load_options (Optional[dict]): A dictionary of options to pass to the checkpoint loading function.
        event (Union[str, Event]): The event at which to load the checkpoint. Defaults to ``Event.BEFORE_LOAD``.
    """

    def __init__(
        self,
        load_path: str,
        load_weights_only: bool = False,
        strict_model_weights: bool = True,
        ignore_keys: Optional[list[str]] = None,
        event: Union[str, Event] = Event.BEFORE_LOAD,
    ):
        super().__init__()
        self.load_path = load_path
        self.load_object_store = maybe_create_object_store_from_uri(load_path)
        _, _, self.parsed_path = parse_uri(load_path)

        self.load_weights_only = load_weights_only
        self.strict_model_weights = strict_model_weights
        self.ignore_keys = ignore_keys

        self.event = event if isinstance(event, Event) else Event[event.upper()]

    def run_event(self, event: Event, state: State, logger: Logger) -> None:
        if event == self.event:
            log.info(f'Loading checkpoint from {self.load_path} at {self.event}.')
            self._load(state, logger)
            log.info(f'Finished loading checkpoint from {self.load_path} at {self.event}.')

        return super().run_event(event, state, logger)

    def _load(self, state: State, logger: Logger) -> None:

        # We need to temporarily disable the `should_save_peft_only` flag on the model
        # so that we can have access to the full model weights for loading.
        model = state.model
        original_should_save_peft_only = False
        if isinstance(model, HuggingFaceModel):
            original_should_save_peft_only = model.should_save_peft_only
            model.should_save_peft_only = False

        load_checkpoint(
            path=self.parsed_path,
            state=state,
            logger=logger,
            object_store=self.load_object_store,
            strict_model_weights=self.strict_model_weights,
            ignore_keys=self.ignore_keys,
            load_weights_only=self.load_weights_only,
        )

        # Restore the original `should_save_peft_only` flag on the model
        if isinstance(model, HuggingFaceModel):
            model.should_save_peft_only = original_should_save_peft_only



# Copyright 2024 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""Useful functions for load checkpoints from remote object store or local disk."""

import logging
from typing import Optional

from composer.utils import (
    dist,
    extract_path_from_symlink,
    maybe_create_object_store_from_uri,
    parse_uri,
    retry,
)

log = logging.getLogger(__name__)


def download_file(
    source_uri: str,
    destination_path: str,
    node_ranks: Optional[list[int]] = None,
    num_attempts: int = 5,
):
    """Downloads a file (object) from the specified URI to the specified directory.

    Args:
        source_uri (str): The URI to download the file from or a symlink to the URI.
        destination_path (str): The directory to download the file to.
        node_ranks (list[int]): The ranks of the nodes that will download the file. If None, all nodes will download the file.
        num_attempts (int): Retry for object store downloads. Default to 5.
    """
    # Only local rank 0 downloads
    local_rank = dist.get_local_rank()
    if local_rank != 0:
        return

    node_rank = dist.get_node_rank()
    if node_ranks is not None and node_rank not in node_ranks:
        return

    if source_uri.endswith('.symlink'):
        source_path = download_and_extract_symlink(source_uri)

    object_store = maybe_create_object_store_from_uri(source_uri)
    _, _, source_path = parse_uri(source_uri)
    assert object_store is not None

    @retry(num_attempts=num_attempts)
    def _download():
        object_store.download_object(
            object_name=source_path,
            filename=destination_path,
        )

    log.debug(f'Downloading {source_path} to {destination_path}')
    _download()
    log.debug(f'Finished downloading {source_path} to {destination_path}')


def download_and_extract_symlink(source_uri) -> str:
    """Downloads a symlink file from the specified URI and returns the path it points to.

    Args:
        source_uri (str): The URI to download the symlink from.

    Returns:
        str: The path the symlink points to.
    """
    object_store = maybe_create_object_store_from_uri(source_uri)
    _, _, source_path = parse_uri(source_uri)
    source_path = extract_path_from_symlink(source_path, object_store)
    return source_path


def download_monolithic_checkpoint(
    source_uri: str,
    destination_path: str,
    global_rank_zero_only: bool = True,
):
    """Downloads a monolithic checkpoint from the specified URI to the specified directory.

    Args:
        source_uri (str): The URI to download the checkpoint from or symlink that points to the URI.
        destination_path (str): The directory to download the checkpoint to.
        global_rank_zero_only (bool): If True, only rank 0 will download the checkpoint.
        broadcast_files_to_other_nodes (bool): If True, the downloaded checkpoint will be broadcast to all other nodes.
            If torch syncs modules states this is unnecessary.
    """
    node_ranks = None
    if global_rank_zero_only:
        node_ranks = [0]
    download_file(
        source_uri=source_uri,
        destination_path=destination_path,
        node_ranks=node_ranks,
    )
    if not global_rank_zero_only or (global_rank_zero_only and dist.get_global_rank() == 0):
        return destination_path
    return None



# Copyright 2024 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""Useful functions for saving state dicts to disk."""

import json
import logging
import os
import pickle
import textwrap
import warnings
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Optional, Sequence, Union

import torch
import torch.distributed.checkpoint as DCP
from packaging import version
from torch.distributed._shard.sharded_tensor import ShardedTensor
from torch.distributed._tensor import DTensor

from composer.checkpoint.state_dict import (
    get_metadata_state_dict,
    get_model_state_dict,
    get_optim_state_dict,
    get_resumption_state_dict,
)
from composer.core import State, Time
from composer.devices import Device
from composer.models import ComposerModel
from composer.utils import dist
from composer.utils.checkpoint import _TORCH_DISTRIBUTED_CHECKPOINTS_FILENAME, _write_checkpoint_file
from composer.utils.file_helpers import format_name_with_dist_and_time

log = logging.getLogger(__name__)

MODEL_CHECKPOINT_DIRECTORY_NAME = 'model'
MONOLITHIC_MODEL_CHECKPOINT_FILENAME = 'model.pt'
OPTIM_CHECKPOINT_DIRECTORY_NAME = 'optim'
OPTIM_MONO_CHECKPOINT_FILENAME = 'optim.pt'
METADATA_CHECKPOINT_FILENAME = 'composer_metadata.json'
RESUMPTION_CHECKPOINT_FILENAME = 'resumption.pkl'


@dataclass
class CheckpointSaveOptions:
    """Options for saving a checkpoint to disk.

    Args:
        destination_dir (str): The directory to save the checkpoint to.
        save_frequency (Union[str, int, Time]): The frequency to save the checkpoint.
            If '1ep', the checkpoint will be saved after each epoch.
            If '1ba', the checkpoint will be saved after each batch.
            If an int, the checkpoint will be saved after that many epochs.
        dir_prefix (str): The prefix to use for the directory name. Can include {epoch} and {batch}.
        overwrite (bool): Whether to overwrite the checkpoint if it already exists.
        save_model (bool): Whether to save the model.
        save_optimizer (bool): Whether to save the optimizer.
        save_resumption_state (bool): Whether to save the resumption state.
        num_checkpoints_to_keep (int): The number of checkpoints to keep.
            If -1, all checkpoints will be kept.
        save_format (str): The format to save the model in. 'pt', which is the standard pytorch serializarion, is the only option for now.
        sharded_checkpoint (bool): Whether to save the model as a sharded checkpoint.
        precision (str): The precision to save the model in. One of 'bf16', 'fp32', 'fp16', 'fp64'.
        include_keys (Optional[Union[str, Sequence[str]]]): Keys to include in the saved model.
        ignore_keys (Optional[Union[str, Sequence[str]]]): Keys to ignore in the saved model.
    """
    save_frequency: Union[str, int, Time] = '1ep'
    dir_prefix: str = 'ep{epoch}-ba{batch}'
    overwrite: bool = False
    save_model: bool = True
    save_optimizer: bool = True
    save_resumption_state: bool = True
    num_checkpoints_to_keep: int = -1
    save_format: str = 'pt'
    sharded_checkpoint: bool = False
    precision: str = 'fp32'
    include_keys: Optional[Union[str, Sequence[str]]] = None
    ignore_keys: Optional[Union[str, Sequence[str]]] = None


def save_checkpoint_to_disk(
    destination_dir: str,
    state: State,
    options: Optional[Union[CheckpointSaveOptions, dict]] = None,
):
    """Saves a checkpoint to disk.

    Args:
        state (State): The state to save.
        options (Optional[Union[CheckpointSaveOptions, Dict]]): The options for saving the checkpoint.
            If None, destination_dir must be provided.
        destination_dir (Optional[str]): The directory to save the checkpoint to.
            If options is provided, this will overwrite save_path.
    """
    if options is None:
        options = CheckpointSaveOptions()
    else:
        if isinstance(options, dict):
            options = CheckpointSaveOptions(**options)

    save_path = os.path.join(destination_dir, options.dir_prefix)
    save_path = format_name_with_dist_and_time(save_path, state.run_name, state.timestamp)
    os.makedirs(save_path, exist_ok=True)
    if options.save_model:
        model_save_path = (
            os.path.join(save_path, MODEL_CHECKPOINT_DIRECTORY_NAME) if options.sharded_checkpoint else
            os.path.join(save_path, MODEL_CHECKPOINT_DIRECTORY_NAME, MONOLITHIC_MODEL_CHECKPOINT_FILENAME)
        )
        save_model_to_disk(
            state.model,
            model_save_path,
            options.sharded_checkpoint,
            options.precision,
            options.include_keys,
            options.ignore_keys,
            options.overwrite,
            options.save_format,
        )
    if options.save_optimizer:
        optim_save_path = os.path.join(save_path,
                                       OPTIM_CHECKPOINT_DIRECTORY_NAME) if options.sharded_checkpoint else os.path.join(
                                           save_path,
                                           OPTIM_CHECKPOINT_DIRECTORY_NAME,
                                           OPTIM_MONO_CHECKPOINT_FILENAME,
                                       )
        optimizer = state.optimizers[0]
        save_optim_to_disk(
            state.model,
            optimizer,
            optim_save_path,
            options.sharded_checkpoint,
            options.precision,
            options.overwrite,
            options.save_format,
        )
    if options.save_resumption_state:
        resumption_save_path = os.path.join(save_path, RESUMPTION_CHECKPOINT_FILENAME)
        save_resumption_state_to_disk(state, resumption_save_path)

    md_save_path = os.path.join(save_path, METADATA_CHECKPOINT_FILENAME)
    save_composer_metadata_to_disk(
        md_save_path,
        state.model,
        options.sharded_checkpoint,
        options.precision,
        state.device,
        state.device_train_microbatch_size,
    )
    return save_path


def save_model_to_disk(
    model: Union[ComposerModel, torch.nn.Module],
    destination_dir: str,
    sharded_checkpoint: bool = False,
    precision: str = 'fp32',
    include_keys: Optional[Union[str, Sequence[str]]] = None,
    ignore_keys: Optional[Union[str, Sequence[str]]] = None,
    overwrite: bool = False,
    save_format: str = 'pt',  # or hf, safetensor
) -> Optional[str]:
    """Saves a model to disk.

    Args:
        model (Union[ComposerModel, torch.nn.Module]): The model to save.
        destination_dir (str): The directory to save the model to.
            Model will be saved as distination_dir/model.pt if sharded_checkpoint is False,
            otherwise all shards will be saved as destination_dir/__<rank>_0.distcp along with a metadata file (destination_dir/.metadata).
        sharded_checkpoint (bool): Whether to save the model as a sharded checkpoint.
        precision (str): The precision to save the model in. One of 'bf16', 'fp32', 'fp16', 'fp64'.
        include_keys (Optional[Union[str, Sequence[str]]]): Keys to include in the saved model.
        ignore_keys (Optional[Union[str, Sequence[str]]]): Keys to ignore in the saved model.
        overwrite (bool): If True, the file will be overwritten if it exists.
        save_format (str): The format to save the model in. One of 'pt', 'hf', or 'safetensor'.

    Returns:
        str: The full path to the saved model.
    """
    if save_format != 'pt':
        raise NotImplementedError(
            f"Saving checkpoint in format {save_format} is not supported. Please choose from ['pt'].",
        )
    model_state_dict = get_model_state_dict(
        model,
        sharded_checkpoint,
        precision,
        include_keys,
        ignore_keys,
    )
    saved_path = save_state_dict_to_disk(
        state_dict=model_state_dict,
        destination_file_path=destination_dir,
        overwrite=overwrite,
        save_format=save_format,
    )
    return saved_path


def save_optim_to_disk(
    model: Union[ComposerModel, torch.nn.Module],
    optimizer: torch.optim.Optimizer,
    destination_dir: str,
    sharded_checkpoint: bool = False,
    precision: str = 'fp32',
    overwrite: bool = False,
    save_format: str = 'pt',
) -> Optional[str]:
    """Saves an optimizer to disk.

    Args:
        model (Union[ComposerModel, torch.nn.Module]): The model to save.
        optimizer (torch.optim.Optimizer): The optimizer to save.
        destination_dir (str): The directory to save the optimizer to.
            Optimizer will be saved as destination_dir if sharded_checkpoint is False,
            otherwise all shards will be saved as destination_dir/__<rank>_0.distcp along with a metadata file (destination_dir/.metadata).
        sharded_checkpoint (bool): Whether to save the optimizer as a sharded checkpoint.
        precision (str): The precision to save the optimizer in. One of 'bf16', 'fp32', 'fp16', 'fp64'.
        overwrite (bool): If True, the file will be overwritten if it exists.
        save_format (str): The format to save the optimizer in. One of 'pt'.
    """
    optim_state_dict = get_optim_state_dict(
        model,
        optimizer,
        sharded_state_dict=sharded_checkpoint,
        precision=precision,
    )
    saved_path = save_state_dict_to_disk(
        state_dict=optim_state_dict,
        destination_file_path=destination_dir,
        overwrite=overwrite,
        save_format=save_format,
    )

    return saved_path


def save_composer_metadata_to_disk(
    destination_file_path: str,
    model: Optional[Union[ComposerModel, torch.nn.Module]] = None,
    sharded_state_dict: Optional[bool] = None,
    precision: Optional[Union[str, torch.dtype]] = None,
    device: Optional[Device] = None,
    device_train_microbatch_size: Optional[Union[int, float]] = None,
):
    """Saves metadata about the model to disk.

    Args:
        destination_file_path (str): The path to save the metadata to.
        model (Optional[Union[ComposerModel, torch.nn.Module]]): The model to save metadata about.
        sharded_state_dict (Optional[bool]): Whether the model is sharded.
        precision (Optional[Union[str, torch.dtype]]): The precision of the model.
        device (Optional[Device]): The device the model is on.
        device_train_microbatch_size (Optional[Union[int, float]]): The device train microbatch size.
    """
    md_dict = get_metadata_state_dict(
        model,
        sharded_state_dict,
        precision,
        device,
        device_train_microbatch_size,
    )
    os.makedirs(str(Path(destination_file_path).parent), exist_ok=True)

    if dist.get_global_rank() == 0:
        with open(destination_file_path, 'w') as f:
            json.dump(md_dict, f, indent=4)
    return destination_file_path


def save_resumption_state_to_disk(
    state: State,
    destination_file_path: str,
):
    """Saves the resumption state to disk.

    Args:
        state (State): The state to save.
        destination_file_path (str): The path to save the resumption state to.
    """
    resumption_state_dict = get_resumption_state_dict(state)
    os.makedirs(Path(destination_file_path).parent, exist_ok=True)
    with open(destination_file_path, 'wb') as f:
        pickle.dump(resumption_state_dict, f)
    return destination_file_path


from composer.utils import dist
from composer.utils.checkpoint import _TORCH_DISTRIBUTED_CHECKPOINTS_FILENAME, _write_checkpoint_file

log = logging.getLogger(__name__)


def save_state_dict_to_disk(
    state_dict: dict[str, Any],
    destination_file_path: str,
    overwrite: bool = False,
    save_format: str = 'pt',  # or hf, safetensor
) -> Optional[str]:
    """Saves a state dict to local disk.

    Args:
        state_dict (Dict[str,Any]): The state dict to save.
        destination_file_path (str): The path to save the state dict to. If sharded,
          this should be the pth to a directory. Otherwise, it should be a path to a file.
        overwrite (bool): If True, the file will be overwritten if it exists.
        save_format (str): The format to save the state dict in. One of 'pt', 'hf', or 'safetensor'.

    Returns:
        str: The full path to the saved state dict if (sharded is false and rank 0) or if sharded is true, otherwise None.
    """
    if state_dict == {}:
        return None
    if is_state_dict_sharded(state_dict):
        path_saved = _save_sharded_state_dict_to_disk(state_dict, destination_file_path, overwrite, save_format)
    else:
        if dist.get_global_rank() == 0:
            path_saved = _save_full_state_dict_to_disk(state_dict, destination_file_path, overwrite, save_format)
        else:
            path_saved = None

    return path_saved


def _save_sharded_state_dict_to_disk(
    state_dict: dict[str, Any],
    destination_file_path: str,
    overwrite: bool = False,
    save_format: str = 'pt',
) -> Optional[str]:

    if save_format != 'pt':
        raise NotImplementedError(
            f"Saving sharded state dict to disk in format {save_format} is not supported. Please choose from ['pt'].",
        )

    if state_dict == {}:
        return None

    # If user specifies filename instead of directory suffixes, strip them and warn
    if len(Path(destination_file_path).suffixes) > 0:
        stripped_path = _strip_suffixes(destination_file_path)
        warnings.warn(
            textwrap.dedent(
                f"""Sharded checkpoints require a directory path not a file path:
            {destination_file_path} will have its extensions stripped and checkpoints will be saved in {stripped_path}
            as {stripped_path}/{_TORCH_DISTRIBUTED_CHECKPOINTS_FILENAME}""",
            ),
        )
        destination_file_path = stripped_path

    # Wait for all ranks to get here before checking if the directory exists.
    dist.barrier()
    if dist.get_global_rank() == 0 and not overwrite and os.path.exists(destination_file_path):
        raise ValueError(f'Directory {destination_file_path} already exists. Set overwrite=True to overwrite it.')

    log.debug(
        f'Starting saving of sharded state dict to {destination_file_path}/{_TORCH_DISTRIBUTED_CHECKPOINTS_FILENAME}',
    )

    # For 2.3.0 and above you can use checkpoint_id, but this version works the best for all versions
    # of torch (and makes pyright happier) that we support, so we use it for now.
    if version.parse(torch.__version__) < version.parse('2.2.0'):
        DCP.save_state_dict(state_dict=state_dict, storage_writer=DCP.FileSystemWriter(destination_file_path))
    else:
        DCP.save(state_dict=state_dict, storage_writer=DCP.FileSystemWriter(destination_file_path))

    log.debug(
        f'Finished saving of sharded state dict to {destination_file_path}/{_TORCH_DISTRIBUTED_CHECKPOINTS_FILENAME}',
    )
    return destination_file_path + '/' + _TORCH_DISTRIBUTED_CHECKPOINTS_FILENAME


def _save_full_state_dict_to_disk(
    state_dict: dict[str, Any],
    destination_file_path: str,
    overwrite: bool = False,
    save_format: str = 'pt',  # or hf, safetensor
) -> Optional[str]:

    if save_format != 'pt':
        raise NotImplementedError(
            f"Saving full state dict to disk in format {save_format} is not supported. Please choose from ['pt'].",
        )

    if not overwrite and os.path.exists(destination_file_path):
        raise ValueError(f'File {destination_file_path} already exists. Set overwrite=True to overwrite it.')

    if dist.get_global_rank() == 0:
        os.makedirs(os.path.dirname(destination_file_path), exist_ok=True)
        _write_checkpoint_file(state_dict=state_dict, filename=destination_file_path)
        return destination_file_path
    return None


def is_state_dict_sharded(state_dict: dict[str, Any]) -> bool:
    """Determines if the state dict is sharded.

    Args:
        state_dict (Dict[str, Any]): The state dict to check.

    Returns:
        bool: Whether the state dict is sharded.
    """
    for value in state_dict.values():
        if isinstance(value, ShardedTensor) or isinstance(value, DTensor):
            return True
        elif isinstance(value, dict):
            is_sharded = is_state_dict_sharded(value)
            if is_sharded:
                return True
    return False


def _strip_suffixes(path: Union[str, Path]) -> str:
    path = Path(path)
    for _ in path.suffixes:
        path = path.with_suffix('')

    return str(path)
